/*****************************************************************************/
/* fesim.c v2                                                                */
/*                                                                           */
/* SDE Front End filter and FADC simulation program for                      */
/* Pierre Auger Observatory                                                  */
/*                                                                           */ 
/* Christian Grunfeld - Nov 2002                                             */
/* cmg@fisica.unlp.edu.ar                                                    */
/*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#include "fesim.h"


/*****************************************************************************/
/* Upsample function, to get 1nsec sampling time if the input has an integer */
/* multiple of 1nsec. It implements linear interpolation between real samples*/
/*****************************************************************************/
void upsample(double *ind1, double *outd1, 
              double *ind2, double *outd2,
              double *ind3, double *outd3) {

    int i;
    int j;

    for(i=0; i<PULSE_LENGH; i++) {
        *(outd1+(SAMPLE_TIME*i))=*(ind1+i);
        *(outd2+(SAMPLE_TIME*i))=*(ind2+i);
        *(outd3+(SAMPLE_TIME*i))=*(ind3+i);
    }
    for(i=0; i<UP_SIZE; i+=SAMPLE_TIME) {
        for(j=1; j<SAMPLE_TIME; j++) {
            *(outd1+i+j)=*(outd1+i) + j*( *(outd1+i+SAMPLE_TIME) - *(outd1+i) ) / SAMPLE_TIME;
            *(outd2+i+j)=*(outd2+i) + j*( *(outd2+i+SAMPLE_TIME) - *(outd2+i) ) / SAMPLE_TIME;
            *(outd3+i+j)=*(outd3+i) + j*( *(outd3+i+SAMPLE_TIME) - *(outd3+i) ) / SAMPLE_TIME;
        } 
    }
}


/*****************************************************************************/
/* Downsample function,  gives 25 nsec sampling time data like the LS        */
/* It implements random phase shift from the input pulse and FADC            */
/* digitization level outputs                                                */
/*****************************************************************************/
void downsample(double *ind1, int *outd1,
                double *ind2, int *outd2,
                double *ind3, int *outd3) {

    int i;
    int j;

    /* Initial random phase */
    j=(int)(24.1*rand()/(RAND_MAX+1.0));

    for(i=0; i<DOWN_SIZE; i++) {
        *(outd1+i)=(int) (*(ind1+j+(i*25)) * 1023);
        if(*(outd1+i) > 1023) *(outd1+i)=1023;
        if(*(outd1+i) < 0) *(outd1+i)=0;

        *(outd2+i)=(int) (*(ind2+j+(i*25)) * 1023);
        if(*(outd2+i) > 1023) *(outd2+i)=1023;
        if(*(outd2+i) < 0) *(outd2+i)=0;

        *(outd3+i)=(int) (*(ind3+j+(i*25)) * 1023);
        if(*(outd3+i) > 1023) *(outd3+i)=1023;
        if(*(outd3+i) < 0) *(outd3+i)=0;
    }
}


/*****************************************************************************/
/* Filter function, the core of the FE simulator. Receives a 1nsec spaced    */
/* pulse array and returns a 1nsec filtered pulse array. It also receives    */
/* an array containing the recursive equation coefficients                   */
/*****************************************************************************/
void filter(double *ind1, double *outd1,
            double *ind2, double *outd2,
            double *ind3, double *outd3, double *d) {

    double *x1, *x2, *x3;
    double *y1, *y2, *y3;
    
    /* static variables, to mantain the filter charged between pulses */
    static double mem1[5], mem2[5], mem3[5];

    int i;

    x1=ind1;
    x2=ind2;
    x3=ind3;

    y1=(double *)calloc(UP_SIZE+5, sizeof(double));    
    y2=(double *)calloc(UP_SIZE+5, sizeof(double));
    y3=(double *)calloc(UP_SIZE+5, sizeof(double));    

    for(i=0; i<5; i++) {
        y1[i]=mem1[i];
        y2[i]=mem2[i];
        y3[i]=mem3[i];
    }

    /* Filter recursive equation */
    for(i=5; i<UP_SIZE+5; i++) {
        y1[i]= d[0]*x1[i-5] + d[1]*y1[i-1] + d[2]*y1[i-2] + d[3]*y1[i-3] + d[4]*y1[i-4] + d[5]*y1[i-5];
        y2[i]= d[0]*x2[i-5] + d[1]*y2[i-1] + d[2]*y2[i-2] + d[3]*y2[i-3] + d[4]*y2[i-4] + d[5]*y2[i-5];
        y3[i]= d[0]*x3[i-5] + d[1]*y3[i-1] + d[2]*y3[i-2] + d[3]*y3[i-3] + d[4]*y3[i-4] + d[5]*y3[i-5];
    }

    memcpy((void *)outd1, (void *)(y1+5), UP_SIZE*sizeof(double));
    memcpy((void *)outd2, (void *)(y2+5), UP_SIZE*sizeof(double));
    memcpy((void *)outd3, (void *)(y3+5), UP_SIZE*sizeof(double));

    for(i=0; i<5; i++) {
        mem1[i]=*(y1+UP_SIZE+i);
        mem2[i]=*(y2+UP_SIZE+i); 
        mem3[i]=*(y3+UP_SIZE+i); 
    }

    free(y1);
    free(y2); 
    free(y3); 
}    


/*****************************************************************************/
/* Main function of the stand alone FE and LS simulator                      */
/* For non stand alone aplications remove main but mantain variables and     */
/* calls structures                                                          */
/*****************************************************************************/
int main() {

    /* Input, after filter and output files */
    FILE *filein;
    FILE *filemid;
    FILE *fileout;

    char buf1[20];
    char buf2[20];
    char buf3[20];
    int i;
    int j;

    /* Filter transfer function denominator coefficients in descending */
    /* powers of S */
    /* Correspond to a two stages of 2nd and 3rd order Bessel filters */
    /* with gain -1/2 and normalized to fcut=1Hz */
 
    double g=-0.5;
    double a1=0.00002249469490589952;
    double a2=0.00079636432469041;
    double a3=0.01250643955087904;
    double a4=0.1069326885297293;
    double a5=0.4949718730157946;
    double a6=1.0;
 
    /* Auxiliar coeficients */
    double b1, b2, b3, b4, b5, b6;
    double c1, c2, c3, c4, c5, c6;

    /* Difference recursive equation coeficients array*/
    double d[6];
 
    double T, T2, T3, T4, T5;

    /* Input arrays */
    double in1[PULSE_LENGH];
    double in2[PULSE_LENGH];
    double in3[PULSE_LENGH];

    /* Upsampled arrays */ 
    double up1[UP_SIZE];
    double up2[UP_SIZE];
    double up3[UP_SIZE];

    /* Filtered arrays */
    double filt1[UP_SIZE];
    double filt2[UP_SIZE];
    double filt3[UP_SIZE];

    /* Output arrays */
    int out1[DOWN_SIZE];
    int out2[DOWN_SIZE];
    int out3[DOWN_SIZE];

    /* Open files with input data and to store output data */
    filein=fopen("scope.dat","r");
    filemid=fopen("filter.dat","w");
    fileout=fopen("station.dat","w");

    /* Calculate difference equation coefficients */
    b1=1.0;
    b2=a2/a1;
    b3=a3/a1;
    b4=a4/a1;
    b5=a5/a1;
    b6=a6/a1;
 
    /* T is the denormalization coefficient */
    T=0.000000001*19000000;
    T2=T*T;
    T3=T2*T;
    T4=T3*T;
    T5=T4*T;
 
    c1=1 + b2*T + b3*T2 + b4*T3 + b5*T4 + b6*T5;
    c2=5 + 4*b2*T + 3*b3*T2 + 2*b4*T3 + b5*T4;
    c3=10 + 6*b2*T + 3*b3*T2 + b4*T3;
    c4=10 + 4*b2*T + b3*T2;
    c5=5 + b2*T;
    c6=1.0;

    d[0]=g/a1/c1*T5;
    d[1]=c2/c1;
    d[2]=-c3/c1;
    d[3]=c4/c1;
    d[4]=-c5/c1;
    d[5]=c6/c1;

    /* Scan input file to get the scope data */
    for(i=0; i<PULSE_NUM; i++) {
        for(j=0; j<PULSE_LENGH; j++) {
            fscanf(filein, "%s", buf1);
            fscanf(filein, "%s", buf2); 
            fscanf(filein, "%s", buf3);
            in1[j]=atof(buf1);
            in2[j]=atof(buf2);
            in3[j]=atof(buf3); 
        }

        /* Call upsample() to get the upsampled data */
        upsample(in1, up1, in2, up2, in3, up3);

        /* Call filter() to get the filtered data */
        filter(up1, filt1, up2, filt2, up3, filt3, d);

        /* Call downsample() to get the downsampled data like from LS */
        downsample(filt1, out1, filt2, out2, filt3, out3);
        
        /* Save the pulses in files */
        for(j=0; j<UP_SIZE; j++) {
            fprintf(filemid, "%.14f %.14f %.14f\n", filt1[j], filt2[j], filt3[j]);
        }
        for(j=0; j<DOWN_SIZE; j++) {
            fprintf(fileout, "%d %d %d\n", out1[j], out2[j], out3[j]);
        }
    }

    fclose(filein);
    fclose(filemid);
    fclose(fileout);
    return 0;
}

